/*
 * brain.js v1.0
 * (c) 2021 R.A. Perepelkin
 * Under the MIT License.
 */
import '/web/oda/oda.js'
import * as THREE from '../threejs/lib/three.module.js';
import { OrbitControls } from '../threejs/lib/OrbitControls.js';
import { TrackballControls } from '../threejs/lib/TrackballControls.js';
import { Line2 } from '../threejs/lib/lines/Line2.js';
import { LineMaterial } from '../threejs/lib/lines/LineMaterial.js';
import { LineGeometry } from '../threejs/lib/lines/LineGeometry.js';
import SpriteText from '../threejs/lib/three-spritetext.js';

ODA({
    is: 'oda-neural-space',
    template: `
    <style>
        :host {
            @apply --flex;
            postion: relative;
        }
    </style>
        <canvas ref="canvas" class="flex vertical" style="width:100%;height:100%;">
    `,
    onChanges(e) {
        //console.log(e);
    },
    props: {
        net: {
            type: Object,
            set(n, o) {
                if (o)
                    o.unlisten(this.onChanges);
                this.focused = n;
                n?.listen(this.onChanges)
                if (!this.isNet) {
                    this.createHover();
                    this.isNet = true;
                }
            }
        },
        selection: [],
        focused: Object,
        focusedLayer: 0,
        enableShowGrid: {
            type: Boolean,
            default: true,
            set(n) {
                this.setGridXY(this.showGrid);
            }
        },
        showGrid: {
            type: Boolean,
            default: false,
            set(n) {
                this.setGridXY(n);
            }
        },
        rotate: {
            set(n) {
                this.controls.autoRotate = n;
            }
        },
        zoom: 10,
        _width: {
            reactive: true,
            get() {
                return this.canvas?.$domHost?.offsetWidth || window.innerWidth;
            }
        },
        _height: {
            reactive: true,
            get() {
                return this.canvas?.$domHost?.offsetHeight || window.innerHeight;
            }
        },
        aspect: {
            reactive: true,
            get() {
                return this._width / this._height;
            }
        },
        campos: {
            type: Object,
            default: { x: 0, y: 100, z: 0 },
            set(n) {
                this.setCameraPosition();
            }
        },
        camrot: {
            type: Object,
            default: { x: 100, y: 0, z: 100, angle: 0, step: 0.01, auto: false },
            set(n) {
                this.setCameraRotate();
            }
        },
        camview: {
            type: String,
            list: ['top', 'bottom', 'left', 'right', 'front', 'back', 'back-200', 'rotate'],
            set(n) {
                if (n) {
                    const arr = n.split('-');
                    const act = arr[0];
                    const val = arr[1] || 100;
                    const action = {
                        top: () => this.setCameraPosition(0, val, 0),
                        bottom: () => this.setCameraPosition(0, -val, 0),
                        left: () => this.setCameraPosition(-val, 0, 0),
                        right: () => this.setCameraPosition(val, 0, 0),
                        front: () => this.setCameraPosition(0, 0, val),
                        back: () => this.setCameraPosition(0, 0, -val),
                        rotate: () => this.camrot = { x: val, y: 0, z: val, angle: 0, step: -0.01, auto: true }
                    }
                    if (action[act]) action[act]();
                }
            }
        },
        showTestElements: {
            type: Boolean,
            default: false,
            set(n) {
                this.setTestElements();
            },
            reflectToAttribute: true
        },
        showAxesXYZ: {
            type: Boolean,
            default: false,
            set(n) {
                this.setAxesXYZ(n);
            }
        }
    },
    setAxesXYZ(show) {
        if (this._axesXYZ) {
            this.group.remove(this._axesXYZ);
            delete this._axesXYZ;
            this.group.remove(this._groupLabel);
            delete this._groupLabel;
        }
        if (!show) return;
        const length = 500;
        const material = new THREE.LineBasicMaterial({ color: 0xffff00 });
        const points = [];
        points.push(new THREE.Vector3(0, 0, 0));
        points.push(new THREE.Vector3(0, length, 0));
        points.push(new THREE.Vector3(0, 0, 0));
        points.push(new THREE.Vector3(length, 1, 0));
        points.push(new THREE.Vector3(0, 0, 0));
        points.push(new THREE.Vector3(0, 1, length));
        const geometry = new THREE.BufferGeometry().setFromPoints(points);
        this._axesXYZ = new THREE.Line(geometry, material);
        this.group.add(this._axesXYZ);

        this._groupLabel = new THREE.Group();
        const dist = 13, size = 3;
        const labels = [{ name: 'x', x: dist, y: 0, z: 0 }, { name: 'y', x: 0, y: dist, z: 0 }, { name: 'z', x: 0, y: 0, z: dist }];
        labels.forEach(i => {
            const l = new SpriteText(i.name, size);
            l.color = 'orange';
            l.position.set(i.x, i.y, i.z);
            this._groupLabel.add(l);
        })
        this.group.add(this._groupLabel);
    },
    setTestElements() {
        if (this._groupTest) {
            this._groupTest.remove(this._testEl_1);
            delete this._testEl_1;
            this._groupTest.remove(this._testEl_2);
            delete this._testEl_2;
            this._groupTest.remove(this._testEl_3);
            delete this._testEl_3;
            this.group.remove(this._groupTest);
            delete this._groupTest;
            this.showAxesXYZ = false;
            this.showGrid = false;
        }
        if (!this.showTestElements) return;
        this._groupTest = new THREE.Group();
        this._testEl_1 = new THREE.Mesh(new THREE.BoxGeometry(1 * this.zoom, 1 * this.zoom, 1 * this.zoom), new THREE.MeshBasicMaterial({ color: 0xff00ff, wireframe: true }));
        this._testEl_1.position.set(0, this.zoom / 2 + 0.1, 0);
        this._groupTest.add(this._testEl_1);
        this._testEl_2 = new THREE.Mesh(new THREE.SphereGeometry(4, 12, 12), new THREE.MeshBasicMaterial({ color: 0xff0000, wireframe: true }));
        this._testEl_2.position.set(this.zoom * 3, 0, 0);
        this._groupTest.add(this._testEl_2);
        this._testEl_3 = new THREE.Mesh(new THREE.OctahedronGeometry(4), new THREE.MeshBasicMaterial({ color: 0x00ff00, wireframe: true }));
        this._testEl_3.position.set(0, this.zoom * 2, -this.zoom * 2);
        this._groupTest.add(this._testEl_3);
        this.group.add(this._groupTest);
        this.showGrid = true;
        this.showAxesXYZ = true;
        this.setAxesXYZ(true);
    },
    setCameraPosition(x = this.campos.x, y = this.campos.y, z = this.campos.z) {
        if (!this.camera) return;
        this.camrot.auto = false;
        this.camera.position.set(x, y, z);
    },
    setCameraRotate(x = this.camrot.x, y = this.camrot.y, z = this.camrot.z, step = this.camrot.step, auto = this.camrot.auto) {
        if (!this.camera) return;
        if (auto) this.camrot.angle += step;
        if (x) this.camera.position.x = x * Math.cos(this.camrot.angle);
        if (y) this.camera.position.y = y * Math.sin(this.camrot.angle);
        if (z) this.camera.position.z = z * Math.sin(this.camrot.angle);
    },
    listeners: {
        resize: function () {
            this.resize();
        }
    },
    attached() {
        this.initThree();
        window.addEventListener('resize', this.resize.bind(this), false)
    },
    initThree() {
        this.canvas = this.$refs.canvas;
        this.renderer = new THREE.WebGLRenderer({ canvas: this.canvas, antialias: true });
        this.renderer.setPixelRatio(window.devicePixelRatio);
        this.renderer.setSize(this._width, this._height);
        // this.renderer.physicallyCorrectLights = true;

        this.scene = new THREE.Scene();
        this.scene.background = new THREE.Color('black');
        this.group = new THREE.Group();
        this.scene.add(this.group);

        // this.camera = new THREE.PerspectiveCamera(45, this.aspect, .1, 4000);
        // this.setCameraPosition();
        this.camera = new THREE.PerspectiveCamera(45, this.aspect, .1, 4000);
        this.camera.position.z = -100;
        this.camera.position.y = -100;
        this.camera.position.x = -50;

        this.controls = new OrbitControls(this.camera, this.canvas);
        this.controls.enableDamping = true;
        this.controls.dampingFactor = 0.05;
        // this.controls.listenToKeyEvents(window);
        // this.controls.maxPolarAngle = Math.PI / 2;
        // this.controls.enableZoom = true;
        // this.controls.enablePan = true;

        // this.controls = new TrackballControls(this.camera, this.canvas);
        // this.controls.rotateSpeed = 2;
        // this.controls.zoomSpeed = 2;
        // this.controls.panSpeed = 2;

        this.setTestElements();

        this.setGridXY();

        this.loop();
    },
    resize() {
        this.camera.aspect = this._width / this._height;
        this.camera.updateProjectionMatrix();
        this.renderer.setSize(this._width, this._height);
    },
    setGridXY(show = this.showGrid) {
        if (this._gridXY) {
            this._gridXY && this.group.remove(this._gridXY);
            delete this._gridXY;
        }
        if (!(this.enableShowGrid && show)) return;
        this._gridXY = new THREE.GridHelper(1000, 200, 'gray', 'rgba(80,80,80)');
        this.group.add(this._gridXY);
        this._gridXY.position.y = (this.focusedLayer || 0) * this.zoom;
    },
    createHover() {
        this.raycaster = new THREE.Raycaster();
        this.mouse = new THREE.Vector2();
        this.canvas.addEventListener('mousemove', this.onMouseMove.bind(this), false);
        this.canvas.addEventListener('click', this.onMouseDown.bind(this), true);
    },
    loop() {
        requestAnimationFrame(() => this.loop());
        this.controls.update();
        this.renderer.render(this.scene, this.camera);
        if (this.isNet) {
            this.update();
            this.checkHover();
        }
        if (this.camrot.auto)
            this.setCameraRotate();
    },
    update() {
        this.modulation = this.net.modulation;
        for (let l of Object.values(this.net.layers))
            for (let n of l.neurons) {
                this.updatePoint(n);
                if (this.hoverItem === n || this.selection.includes(n)) {
                    n.threePoint.material.color.setHex(0xffffff);
                    this.updateLine(n);
                }
                else
                    this.removeLine(n);
            }
        this.updateLayer(this.net);
        //this.canvas.$domHost.render();
    },
    updateLayer(n) {
        if (n.threePoint) {
            n.threePoint.scale.x = n.threePoint.scale.y = n.threePoint.scale.z = 0.5;
            n.threePoint.material.color.setHex(0xff8000);
            return;
        }
        this.createPoint(n, 0xff8000, 0.5);
    },
    updatePoint(n) {
        let color = 0x0000ff;
        if (n.inputs?.length) {
            color = !n.charge ? 0x888888 : n.charge > 0 ? 0x22ff22 : 0xff3333;
        }
        const scale = (Math.sqrt(Math.abs(n.charge)) + 10) / 10;
        if (n.threePoint) {
            n.threePoint.scale.x = n.threePoint.scale.y = n.threePoint.scale.z = scale;
            n.threePoint.material.color.setHex(color);
            return;
        }
        this.createPoint(n, color, scale);
    },
    createPoint(n, color, scale) {
        const geometry = new THREE.SphereGeometry(1, 32, 32);
        const material = new THREE.MeshBasicMaterial({ color });
        const point = new THREE.Mesh(geometry, material);
        point.scale.x = point.scale.y = point.scale.z = scale;
        point.position.x = n.x * this.zoom;
        point.position.y = n.z * this.zoom;
        point.position.z = n.y * this.zoom;
        this.group.add(point);
        n.threePoint = point;
        point.nnItem = n;
        n._removePoint = (n) => { this.group.remove(n.threePoint) };
    },
    removeLine(n) {
        for (let s of n?.inputs || []) {
            if (!s.threeLine) {
                continue;
            }
            this.group.remove(s.threeLine);
            s.threeLine = null;
        }
    },
    updateLine(n) {
        for (let s of n?.inputs || []) {
            if (s.threeLine) {
                // s.threeLine.material.linewidth = Math.sqrt(s.strong) || 1;
                if (s.w > 0)
                    s.threeLine.material.color.setRGB(0, s.w * 2, 0);
                else if (s.w < 0)
                    s.threeLine.material.color.setRGB(-s.w * 2, 0, 0);
                else
                    s.threeLine.material.color.setRGB(.5, .5, .5);
                continue;
            }
            const points = [new THREE.Vector3(s.neuron.x * this.zoom, s.neuron.z * this.zoom, s.neuron.y * this.zoom)];
            points.push(new THREE.Vector3(s.source.x * this.zoom, s.source.z * this.zoom, s.source.y * this.zoom));

            s.threeLine = this.makeLine(this.group, points);
            s.threeLine.nnItem = s;
        }
    },
    makeLine(group, points, c = 'lightgray', w = 1, d = 1) {
        const point = new THREE.Vector3();
        let color = new THREE.Color();
        const positions = [];
        const colors = [];
        const spline = new THREE.CatmullRomCurve3(points);
        const divisions = Math.round(d * points.length);
        let fnC = () => color = c;
        if (c === 'rainbow') fnC = (color, t) => color.setRGB(0, 0, 0);//setHSL(t, 0, 0.5);
        else if (typeof c == 'string') fnC = (color) => color.set(c);
        for (let i = 0, l = divisions; i <= l; i++) {
            const t = i / l;
            spline.getPoint(t, point);
            positions.push(point.x, point.y, point.z);
            fnC(color, t);
            colors.push(color.r, color.g, color.b);
        }
        const geometry = new LineGeometry();
        geometry.setPositions(positions);
        geometry.setColors(colors);
        const matLine = new LineMaterial({
            color: 0x000000,
            linewidth: w,
            vertexColors: true,
            resolution: new THREE.Vector2(this._width, this._height),
            dashed: false
        });
        const line = new Line2(geometry, matLine);
        line.computeLineDistances();
        line.scale.set(1, 1, 1);
        group.add(line);
        return line;
    },
    checkHover() {
        this.camera.updateMatrixWorld();
        this.raycaster.setFromCamera(this.mouse, this.camera);
        this.intersects = this.raycaster.intersectObjects(this.scene.children[0].children);
        this.hoverItem = undefined;
        for (let i = 0; i < this.intersects.length; i++)
            if (this.intersects[i].object.nnItem) {
                this.intersects[i].object.material.color.set(0xffff00);
                this.hoverItem = this.intersects[i].object.nnItem;
                return;
            }
    },
    onMouseMove(e) {
        // e.preventDefault();
        const rect = this.canvas.getBoundingClientRect();
        const x = (e.clientX - rect.left) * this.canvas.width / rect.width;
        const y = (e.clientY - rect.top) * this.canvas.height / rect.height;
        this.mouse.x = (x / this.canvas.width) * 2 - 1;
        this.mouse.y = (y / this.canvas.height) * -2 + 1;
    },
    onMouseDown(e) {
        if (this.hoverItem) {
            if (this.enableShowGrid && e.ctrlKey) {
                if (this.focusedLayer === this.hoverItem.layer.index && this.showGrid) {
                    this.showGrid = !this.showGrid;
                    this.setGridXY();
                    return;
                }
                this.setGridXY(false);
                this.focusedLayer = this.hoverItem.layer.index;
                this.setGridXY(true);
                this.showGrid = true;
            } else {
                this.focused = this.hoverItem;
                const idx = this.selection.indexOf(this.hoverItem);
                if (~idx)
                    this.selection.splice(idx, 1);
                else
                    this.selection.add(this.hoverItem);
            }
        } else {
            this.focused = this.net;
            this.selection.splice();
            if (this.enableShowGrid && e.ctrlKey) {
                this.focusedLayer = 0;
                this.showGrid = !this.showGrid;
                this.setGridXY();
            }
        }
    }
})