/*
 * brain.js v1.0
 * (c) 2021 R.A. Perepelkin
 * Under the MIT License.
 */

import '/web/oda/components/layouts/app-layout/app-layout.js';
import '/web/oda/tools/property-grid/property-grid.js';
import './space.js'
import NeuroNet from './neuronet.js';

import * as THREE from './threejs/lib/three.module.js';
import { OrbitControls } from './threejs/lib/OrbitControls.js';

import { Line2 } from './threejs/lib/lines/Line2.js';
import { LineMaterial } from './threejs/lib/lines/LineMaterial.js';
import { LineGeometry } from './threejs/lib/lines/LineGeometry.js';

import RendererStats from './threejs/lib/threex.rendererstats.js';

import SpriteText from './threejs/lib/three-spritetext.js';

class cortexMesh {
    zoom = 10
    constructor(nnProps = {}, canvas, brain) {
        this.brain = brain;
        this.canvas = canvas;
        this.nn = new NeuroNet(nnProps.data);
        Object.keys(nnProps.sets || []).forEach(k => this.nn[k] = nnProps.sets[k]);
        this.nnProps = nnProps;
        this.init();
    }
    init() {
        this.nn.listen('removeSynapse', (e) => {
            this.group.remove(e.detail.value.threeLine);
            delete e.detail.value.threeLine;
        })
        this.nn.listen('removeNeuron', (e) => {
            this.group.remove(e.detail.value.threePoint);
            delete e.detail.value.threePoint;
        })
        window.addEventListener('resize', this.resize, false);
        this.createRenderer();
        this.createScene();
        this.createCamera();
        //this.createLights();
        this.createControls();
        this.createGridXY();
        this.createStats();
        this.createLabels();
        this.createHover();
        this.loop();
    }
    resize = () => {
        this.camera.aspect = window.innerWidth / window.innerHeight;
        this.camera.updateProjectionMatrix();
        this.renderer.setSize(window.innerWidth, window.innerHeight);
    }
    createRenderer() {
        this.renderer = new THREE.WebGLRenderer({ canvas: this.canvas, antialias: true });
        this.renderer.setPixelRatio(window.devicePixelRatio);
        this.renderer.setSize(window.innerWidth, window.innerHeight);
        this.renderer.physicallyCorrectLights = true;
    }
    createScene() {
        this.scene = new THREE.Scene();
        this.scene.background = new THREE.Color('black');
        this.group = new THREE.Group();
        this.scene.add(this.group);
    }
    createCamera() {
        const aspect = window.innerWidth / window.innerHeight;
        this.camera = new THREE.PerspectiveCamera(45, aspect, 1, 4000);
        this.camera.position.x = -10;
        this.camera.position.y = 100;
        this.camera.position.z = -200;
    }
    createLights() {

    }
    createControls() {
        this.controls = new OrbitControls(this.camera, this.canvas);
        this.controls.enableDamping = true;
        this.controls.dampingFactor = 0.05;
        this.controls.listenToKeyEvents(window);
        this.controls.maxPolarAngle = Math.PI / 2;
    }
    createGridXY() {
        const gridXY = new THREE.GridHelper(1000, 100, 'gray', 'rgba(80,80,80)');
        this.group.add(gridXY);
    }
    createStats() {
        if (!this.showStats) {
            if (this.rendererStats) document.body.removeChild(this.rendererStats.domElement);
            return;
        }
        if (!this.rendererStats) {
            this.rendererStats = new RendererStats();
            this.rendererStats.domElement.style.position = 'absolute';
            this.rendererStats.domElement.style.left = '0px';
            this.rendererStats.domElement.style.bottom = '0px';
        }
        document.body.appendChild(this.rendererStats.domElement);
    }
    createLabels() {
        this.nnProps.data.forEach(i => {
            const SimpleText = new SpriteText(i.label, 4);
            SimpleText.color = 'orange';
            SimpleText.position.x = i.x * this.zoom - 2;
            SimpleText.position.y = i.z * this.zoom + 2;
            SimpleText.position.z = i.y * this.zoom - 2;
            this.group.add(SimpleText);
        })
    }
    createHover() {
        this.raycaster = new THREE.Raycaster();
        this.mouse = new THREE.Vector2();
        document.addEventListener('mousemove', this.onMouseMove.bind(this), false);
    }
    loop() {
        requestAnimationFrame(() => this.loop());
        this.controls.update();
        this.renderer.render(this.scene, this.camera);
        this.update();
        this.checkHover();
        if (this.showStats) this.rendererStats.update(this.renderer);
    }
    update() {
        this.modulation = this.nn.modulation;
        for (let l of Object.values(this.nn.layers))
            for (let n of l.neurons) {
                this.updatePoint(n);
                this.updateLine(n);
                this.updateChart(n);
            }
        this.updateLayer(this.nn);
        this.brain.render();
    }
    updateLayer(n) {
        if (n.threePoint) {
            n.threePoint.scale.x = n.threePoint.scale.y = n.threePoint.scale.z = 0.5;
            n.threePoint.material.color.setHex(0xff8000);
            return;
        }
        this.createPoint(n, 0xff8000, 0.5);
    }
    updatePoint(n) {
        const color = !n.charge ? 0x888888 : n.charge > 0 ? 0x22ff22 : 0xff3333;
        const scale = (Math.sqrt(Math.abs(n.charge)) + 5) / 10;
        if (n.threePoint) {
            n.threePoint.scale.x = n.threePoint.scale.y = n.threePoint.scale.z = scale;
            n.threePoint.material.color.setHex(color);
            return;
        }
        this.createPoint(n, color, scale);
    }
    createPoint(n, color, scale) {
        const geometry = new THREE.SphereGeometry(1, 32, 32);
        const material = new THREE.MeshBasicMaterial({ color });
        const point = new THREE.Mesh(geometry, material);
        point.scale.x = point.scale.y = point.scale.z = scale;
        point.position.x = n.x * this.zoom;
        point.position.y = n.z * this.zoom;
        point.position.z = n.y * this.zoom;
        this.group.add(point);
        n.threePoint = point;
        point.nnItem = n;
    }
    updateLine(n) {
        (n && n.synapses || []).forEach(s => {
            const coords = [s.parent.x, s.parent.y, s.parent.z, s.target.x, s.target.y, s.target.z].join();
            if (s.threeLine) {
                s.threeLine.material.linewidth = Math.sqrt(s.strong) || 1;
                if (s.charge/*impulses.length > 0*/)
                    s.threeLine.material.color.setRGB(0, 1, 0);
                else
                    s.threeLine.material.color.setRGB(1, 1, 1);
                return;
            }
            const points = [];
            points.push(new THREE.Vector3(s.parent.x * this.zoom, s.parent.z * this.zoom, s.parent.y * this.zoom));
            if (s.z > 0) {
                points.push(new THREE.Vector3(s.parent.x * this.zoom, (s.parent.z + s.z / this.zoom) * this.zoom, s.parent.y * this.zoom));
                points.push(new THREE.Vector3(s.target.x * this.zoom, (s.target.z - s.z / this.zoom) * this.zoom, s.target.y * this.zoom));
            }
            points.push(new THREE.Vector3(s.target.x * this.zoom, s.target.z * this.zoom, s.target.y * this.zoom));

            const line = this.makeLine(this.group, points);
            s.threeLine = line;
            line.nnItem = s;
        })
    }
    makeLine(group, points, c = 'lightgray', w = 1, d = 6) {
        const point = new THREE.Vector3();
        let color = new THREE.Color();
        const positions = [];
        const colors = [];
        const spline = new THREE.CatmullRomCurve3(points);
        const divisions = Math.round(d * points.length);
        let fnC = () => color = c;
        if (c === 'rainbow') fnC = (color, t) => color.setHSL(t, 1.0, 0.5);
        else if (typeof c == 'string') fnC = (color) => color.set(c);
        for (let i = 0, l = divisions; i < l; i++) {
            const t = i / l;
            spline.getPoint(t, point);
            positions.push(point.x, point.y, point.z);
            fnC(color, t);
            colors.push(color.r, color.g, color.b);
        }
        const geometry = new LineGeometry();
        geometry.setPositions(positions);
        geometry.setColors(colors);
        const matLine = new LineMaterial({
            color: 0xffffff,
            linewidth: w,
            vertexColors: true,
            resolution: new THREE.Vector2(window.innerWidth, window.innerHeight),
            dashed: false
        });
        const line = new Line2(geometry, matLine);
        line.computeLineDistances();
        line.scale.set(1, 1, 1);
        group.add(line);
        return line;
    }
    updateChart(n) {
        if (n.impulses.length > 1) {
            if (this.showCharts && n._3d && n._3d.lastImpulsesLength === n.impulses.length && n._3d.showChart) return;
            if (n._3d?.groupCharts) {
                //if (this.showCharts && n._3d?.showChart) return;
                n._3d.groupCharts.children.forEach(ch => {
                    if (ch.geometry) ch.geometry.dispose();
                    if (ch.material) ch.material.dispose();
                })
                this.group.remove(n._3d.groupCharts);
                delete n._3d.groupCharts;
                n._3d.lastImpulsesLength = 0;
            }
            if (!this.showCharts || !n._3d?.showChart) return;
            n._3d.lastImpulsesLength = n.impulses.length;
            const group = n._3d.groupCharts = new THREE.Group();
            let points = [],
                x = n.x * this.zoom,
                y = n.z * this.zoom,
                z = n.y * this.zoom + .5;
            
            // let color = 'rainbow';
            // let colorXYX = 'red';
            let color = n._3d.color;
            if (!color) {
                const _color = new THREE.Color();
                color = n._3d.color = _color.setHSL(Math.random(), 1.0, 0.5);
            }
            let colorXYX = color;
            this.makeLine(group, [new THREE.Vector3(x, y, z), new THREE.Vector3(x + 30, y, z)], colorXYX, 1, 2);
            this.makeLine(group, [new THREE.Vector3(x, y - 10, z), new THREE.Vector3(x, y + 10, z)], colorXYX, 1, 2);

            n.impulses.forEach(i => {
                let y = this.zoom * (n.z + i / 2);
                points.push(new THREE.Vector3(x += 1, y, z));
            });

            // test sinusoid
            // let _x = x, i = 2, h = 10;
            // points = [new THREE.Vector3(_x, y, z), new THREE.Vector3(_x += i, y + h, z), new THREE.Vector3(_x += i, y, z), new THREE.Vector3(_x += i, y - h, z), new THREE.Vector3(_x += i, y, z), new THREE.Vector3(_x += i, y + h, z), new THREE.Vector3(_x += i, y, z), new THREE.Vector3(_x += i, y - h, z), new THREE.Vector3(_x += i, y, z), new THREE.Vector3(_x += i, y + h, z), new THREE.Vector3(_x += i, y, z), new THREE.Vector3(_x += i, y - h, z), new THREE.Vector3(_x += i, y, z),];

            const line = this.makeLine(group, points, color, 3, 6);
            this.group.add(group);
        }
    }
    checkHover() {
        this.camera.updateMatrixWorld();
        this.raycaster.setFromCamera(this.mouse, this.camera);
        this.intersects = this.raycaster.intersectObjects(this.scene.children[0].children);
        this.hoverItem = undefined;
        for (let i = 0; i < this.intersects.length; i++)
            if (this.intersects[i].object.nnItem) {
                this.intersects[i].object.material.color.set(0xffff00);
                this.hoverItem = this.intersects[i].object.nnItem;
                return;
            }
    }
    onMouseMove(e) {
        e.preventDefault();
        const rect = this.canvas.getBoundingClientRect();
        const x = (e.clientX - rect.left) * this.canvas.width / rect.width;
        const y = (e.clientY - rect.top) * this.canvas.height / rect.height;
        this.mouse.x = (x / this.canvas.width) * 2 - 1;
        this.mouse.y = (y / this.canvas.height) * -2 + 1;
    }
    get keys() {
        return {
            'Minus': () => { this.controls.autoRotate = false },
            'Equal': () => { this.controls.autoRotate = true },
            'NumpadSubtract': () => { this.camera.position.z -= 5 },
            'NumpadAdd': () => { this.camera.position.z += 5 },
            'Backspace': () => { this.nn.modulation = 100 }
        }
    }
}
ODA({
    is: 'oda-brain', extends: 'oda-app-layout',
    template: `
        <div slot="title" class="horizontal flex between" style="align-items:center;">
            <div class="no-flex horizontal">
                <oda-button allow-toggle :toggled="!cortex?.nn?.stop" icon="av:play-arrow" @tap="cortex.nn.stop=!cortex.nn.stop;cortex?.nn.step()"></oda-button>
                <span style="align-self: center">|</span>
                <oda-button :disabled="!cortex?.nn?.stop" @tap="cortex?.nn.step()" icon="av:skip-next"></oda-button>
            </div>
            <div class="flex horizontal center">
                <oda-button class="border modulation" icon="space-sci-fi:space-ship" @tap="cortex.nn.modulation = 100" ~style="{background:'linear-gradient(to top, #f00 '+modulation+'%, transparent 0%);'}"></oda-button>
            </div>
            <div class="no-flex horizontal">
                <oda-button allow-toggle :toggled="cortex?.showStats" icon="icons:info-outline" @tap="cortex.showStats=!cortex.showStats;cortex.createStats()"></oda-button>
                <oda-button allow-toggle :toggled="cortex?.showCharts" icon="office-set:chart" @tap="cortex.showCharts=!cortex.showCharts"></oda-button>
                <oda-button allow-toggle :toggled="cortex?.controls.autoRotate" icon="icons:3d-rotation" @tap="cortex.controls.autoRotate=!cortex.controls.autoRotate"></oda-button>
                <oda-button allow-toggle :toggled="cortex?.nn?.locked" icon="icons:lock" @tap="cortex.nn.locked=!cortex.nn.locked"></oda-button>
            </div>
        </div>
        <div slot="left-drawer" style="padding: 4px">
            <div style="font-size:small;text-align:center">SENSORS</div>
            <div ~for="k in Object.keys(keys)">
                <div ~if="keys[k].sensors && keys[k].value" style="font-size:small;position:absolute;margin: 2px 0 0 2px;z-index:1">{{keys[k].value}}</div>
                <oda-button ~if="keys[k].sensors" ~style="{backgroundColor:keys[k].pressed?'silver':''}" style="border: 1px solid black; border-radius: 5px; margin: 4px 0px; font-size: xx-large; font-weight: bolder;" @tap="signal(keys[k])">{{keys[k].label}}</oda-button>
            </div>
            <oda-button ~if="Object.keys(keys).length" style="border: 1px solid black; border-radius: 5px; margin: 4px 0px; font-size: xx-large; font-weight: bolder;" 
                @tap="Object.keys(keys).forEach(i=>{if(keys[i].fnUp)keys[i].fnUp()})">All</oda-button>
        </div>
        <canvas ref="canvas" slot="main" @resize="cortex?.resize()" class="flex vertical" @tap="tapOnCanvas"></canvas>
        <oda-property-grid slot="right-drawer" :component="focusedItem"></oda-property-grid>
    `,
    props: {
        keys: {},
        nnProps: {},
        cortex: {
            type: Object,
            freeze: true
        },
        iconSize: 32,
        focusedItem: Object,
        modulation: {
            reactive: true,
            get() {
                return this.cortex?.modulation || 0
            }
        }
    },
    attached() {
        this.cortex = new cortexMesh(this.nnProps, this.$refs.canvas, this);
        (this.cortex.nn.layers[0].neurons || []).forEach(n => {
            const arr = this.keys[n.props.code]?.neurons || [];
            this.keys[n.props.code] = {
                sensors: true, label: n.props.label, value: 0, pressed: false, neurons: [...arr, n],
                fn: (obj = this.keys[n.props.code]) => obj.value++,
                fnUp: (obj = this.keys[n.props.code]) => this.signal(obj),
            };
        });
        this.focusedItem = this.cortex.nn;
        window.addEventListener('keydown', e => {
            //console.log(e.code);
            if (this.keys[e.code] || this.cortex.keys[e.code]) {
                this.keys[e.code] = this.keys[e.code] || {};
                this.keys[e.code].pressed = true;
                this.press();
            }
        }, true);
        window.addEventListener('keyup', e => {
            if (this.keys[e.code] || this.cortex.keys[e.code]) {
                this.keys[e.code].pressed = false;
                if (this.keys[e.code].fnUp) this.keys[e.code].fnUp();
            }
        }, true);
    },
    press() {
        if (this.isPressed) return;
        Object.keys(this.keys).forEach(k => {
            if (this.keys[k].pressed)
                if (this.keys[k].fn) {
                    this.keys[k].fn();
                    this.isPressed = true;
                } else if (this.cortex.keys[k]) {
                    this.cortex.keys[k]();
                    this.isPressed = true;
                }
        });
        if (!this.isPressed) return;
        requestAnimationFrame(() => {
            this.isPressed = false;
            this.press();
        });
    },
    signal(obj) {
        if (!obj) return;
        (obj.neurons || []).forEach(n => n.signal(obj.value || 1));
        obj.value = 0;
    },
    tapOnCanvas() {
        this.focusedItem = this.cortex.hoverItem || this.cortex.nn;
        if (this.cortex.hoverItem) {
            if (this.cortex.showCharts) {
                this.cortex.hoverItem._3d = this.cortex.hoverItem._3d || {};
                this.cortex.hoverItem._3d.showChart = !this.cortex.hoverItem._3d.showChart;
            }
            if (this.cortex.nn.collect[this.cortex.hoverItem.id]) {
                delete this.cortex.nn.collect[this.cortex.hoverItem.id];
            } else {
                this.cortex.nn.collect[this.cortex.hoverItem.id] = [];
            }
        }
    }
})
